@@ -13,7 +13,7 @@ import org.gcube.common.portal.PortalContext;
 import org.gcube.common.scope.api.ScopeProvider;
 import org.gcube.datacatalogue.ckanutillibrary.server.DataCatalogue;
 import org.gcube.datacatalogue.ckanutillibrary.server.DataCatalogueFactory;
-import org.gcube.datacatalogue.ckanutillibrary.server.utils.UtilMethods;
+import org.gcube.datacatalogue.ckanutillibrary.server.utils.CatalogueUtilMethods;
 import org.gcube.datacatalogue.common.Constants;
 import org.gcube.datacatalogue.common.enums.Product_Type;
 import org.gcube.datacatalogue.common.enums.Sources;
@@ -57,13 +57,13 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 	 * Since it needs the scope, we need to check if it is null or not
 	 * @param discoverScope if you want to the discover the utils library in this specified scope
 	 * @return DataCatalogue object
-	 * @throws Exception 
+	 * @throws Exception
 	 */
 	public DataCatalogue getCatalogue(String discoverScope) throws Exception{
 		String currentScope = Utils.getCurrentContext(getThreadLocalRequest(), true);
 		DataCatalogue instance = null;
 		try{
-			String scopeInWhichDiscover = (discoverScope != null && !discoverScope.isEmpty()) ? discoverScope : currentScope;
+			String scopeInWhichDiscover = discoverScope != null && !discoverScope.isEmpty() ? discoverScope : currentScope;
 			logger.debug("Discovering ckan utils library into scope " + scopeInWhichDiscover);
 			instance = DataCatalogueFactory.getFactory().getUtilsPerScope(scopeInWhichDiscover);
 		}catch(Exception e){
@@ -74,16 +74,14 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 	}
 
 	@Override
-	public ManageProductBean getProductBeanById(String productIdentifier) throws Exception {
+	public ManageProductBean getProductBeanById(String productIdentifier, boolean requestForRevertingMerge) throws Exception {
 
 		ManageProductBean toReturn = null;
 
 		// check into user's session first
 		HttpSession httpSession = getThreadLocalRequest().getSession();
-		String sessionProductKey = ScopeProvider.instance.get() + productIdentifier;
-		if(httpSession.getAttribute(sessionProductKey) != null)
-			return (ManageProductBean) httpSession.getAttribute(sessionProductKey);
 
+		// testing case...
 		if(!Utils.isIntoPortal()){
 
 			Thread.sleep(2000);
@@ -100,7 +98,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-of-a-connected-bean",
 					"semantic identifier of the record",
 					"Fishery"
-					));	
+					));
 			toReturn.setCurrentConnections(connectTo);
 
 			// these are the "suggested connections"
@@ -116,7 +114,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 					));
 			toReturn.setSuggestedByKnowledgeBaseConnections(suggestionsForConnections);
 			toReturn.setDomain("Stock");
-			toReturn.setGrsfType("Assessment Unit");
+			toReturn.setCurrentGrsfType("Assessment Unit");
 			toReturn.setKnowledgeBaseId("91f1e413-dc9f-3b4e-b1c5-0e8560177253");
 			toReturn.setShortName("Widow rockfish - US West Coast");
 			toReturn.setShortNameUpdated("Widow rockfish - US West Coast");
@@ -131,55 +129,55 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 			toReturn.setSources(sources);
 			List<SimilarGRSFRecord> similarGrsfRecords = new ArrayList<SimilarGRSFRecord>();
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-1", 
+					"uuid-similar-record-1",
 					"description similar record",
 					"short name similar record 1",
-					"title similar record 1",  
+					"title similar record 1",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-1",
 					"semantic identifier record 1",
 					"Stock 1"
 					));
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-2", 
+					"uuid-similar-record-2",
 					"description similar record",
 					"short name similar record 2",
-					"title similar record 2",  
+					"title similar record 2",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-2",
 					"semantic identifier record 2",
 					"Stock 2"
 					));
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-3", 
+					"uuid-similar-record-3",
 					"description similar record",
 					"short name similar record 3",
-					"title similar record 3",  
+					"title similar record 3",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-3",
 					"semantic identifier record 3",
 					"Stock 3"
 					));
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-4", 
+					"uuid-similar-record-4",
 					"description similar record",
 					"short name similar record 4",
-					"title similar record 4",  
+					"title similar record 4",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-4",
 					"semantic identifier record 4",
 					"Stock 4"
 					));
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-5", 
+					"uuid-similar-record-5",
 					"description similar record",
 					"short name similar record 5",
-					"title similar record 5",  
+					"title similar record 5",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-5",
 					"semantic identifier record 5",
 					"Stock 5"
 					));
 			similarGrsfRecords.add(new SimilarGRSFRecord(
-					"uuid-similar-record-6", 
+					"uuid-similar-record-6",
 					"description similar record",
 					"short name similar record 6",
-					"title similar record 6",  
+					"title similar record 6",
 					"http://data.d4science.org/ctlg/GRSF_Admin/uuid-similar-record-6",
 					"semantic identifier record 6",
 					"Stock 6"
@@ -206,22 +204,25 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 				String systemType = extrasAsMap.get(Constants.SYSTEM_TYPE_CUSTOM_KEY);
 				if(systemType == null || systemType.isEmpty() || systemType.equals(Constants.SYSTEM_TYPE_FOR_SOURCES_VALUE))
-					throw new NoGRSFRecordException("This is not a GRSF Record");	
+					throw new NoGRSFRecordException("This is not a GRSF Record");
 
 				boolean isStock = record.getExtrasAsHashMap().get(Constants.DOMAIN_CUSTOM_KEY).contains(Product_Type.STOCK.getOrigName());
 
 				// fetch map for namespaces
-				Map<String, String> fieldsNamespacesMap = 
-						Utils.getFieldToFieldNameSpaceMapping(httpSession, isStock ? 
+				Map<String, String> fieldsNamespacesMap =
+						Utils.getFieldToFieldNameSpaceMapping(httpSession, isStock ?
 								Constants.GENERIC_RESOURCE_NAME_MAP_KEY_NAMESPACES_STOCK : Constants.GENERIC_RESOURCE_NAME_MAP_KEY_NAMESPACES_FISHERY);
 
 				// get extras as pairs
 				List<CkanPair> extrasAsPairs = record.getExtras();
-
 				Map<String, List<String>> extrasWithoutNamespaces = Utils.replaceFieldsKey(extrasAsPairs, fieldsNamespacesMap);
 				String catalogueIdentifier = record.getId();
 				String description = record.getNotes();
 				Status status = Status.fromString(extrasWithoutNamespaces.get(Constants.STATUS_OF_THE_GRSF_RECORD_CUSTOM_KEY).get(0));
+
+				if(status.equals(Status.To_be_Merged) && !requestForRevertingMerge)
+					throw new Exception("The record is locked due to a merge request in progress!");
+
 				String uuidKB = extrasWithoutNamespaces.get(Constants.UUID_KB_CUSTOM_KEY).get(0);
 				String grsfDomain = extrasWithoutNamespaces.get(Constants.DOMAIN_CUSTOM_KEY).get(0);
 				String semanticId = extrasWithoutNamespaces.get(Constants.GRSF_SEMANTIC_IDENTIFIER_CUSTOM_KEY).get(0);
@@ -236,8 +237,12 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 					logger.warn("Unable to fetch traceability flag. Setting it to false", e);
 				}
 
-				if(status.equals(Status.To_be_Merged) || status.equals(Status.Rejected))
-					throw new Exception("Records under merging activity or rejected cannot be managed!");
+				boolean sdgFlag = false;
+				try{
+					sdgFlag = extrasWithoutNamespaces.get(Constants.SDG_FLAG_CUSTOM_KEY).get(0).equalsIgnoreCase("true");
+				}catch(Exception e){
+					logger.warn("Unable to fetch sdg flag. Setting it to false", e);
+				}
 
 				// Get similar GRSF records, if any (each of which should have name, description, url and id(i.e semantic identifier))
 				List<String> similarGrsfRecordsAsStrings = extrasWithoutNamespaces.containsKey(Constants.SIMILAR_GRSF_RECORDS_CUSTOM_KEY) ? extrasWithoutNamespaces.get(Constants.SIMILAR_GRSF_RECORDS_CUSTOM_KEY): null;
@@ -246,7 +251,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 				if(similarGrsfRecordsAsStrings != null && !similarGrsfRecordsAsStrings.isEmpty()){
 					if(!similarGrsfRecordsAsStrings.get(0).equals(Constants.NO_SIMILAR_GRSF_RECORDS)){
 						for (String similarGRSFRecord : similarGrsfRecordsAsStrings) {
-							similarRecords.add(Utils.similarGRSFRecordFromJson(similarGRSFRecord, catalogue, apiKey));
+							similarRecords.add(Utils.similarGRSFRecordFromJson(similarGRSFRecord, catalogue, apiKey, httpSession));
 						}
 					}
 				}
@@ -254,7 +259,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 				logger.debug("SimilarGRSFRecords are " + similarRecords);
 
 				// get connected records (and the proposed ones)
-				List<String> connectedBeanUrls =  
+				List<String> connectedBeanUrls =
 						extrasWithoutNamespaces.containsKey(Constants.CONNECTED_CUSTOM_KEY) ? extrasWithoutNamespaces.get(Constants.CONNECTED_CUSTOM_KEY): null;
 
 						List<ConnectedBean> connectedBeans = new ArrayList<ConnectedBean>(0);
@@ -272,11 +277,11 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 						// get the connections the knowledge base suggests
 						List<ConnectedBean> suggestedConnectionsByKnowledgeBase = new ArrayList<ConnectedBean>(0);
-						List<String> exploitedResourcesUrls =  isStock ? 
-								(extrasWithoutNamespaces.containsKey(Constants.EXPLOITING_FISHERY_CUSTOM_KEY) ? 
-										extrasWithoutNamespaces.get(Constants.EXPLOITING_FISHERY_CUSTOM_KEY) : null): 
-											(extrasWithoutNamespaces.containsKey(Constants.RESOURCES_EXPLOITED_CUSTOM_KEY) ?
-													extrasWithoutNamespaces.get(Constants.RESOURCES_EXPLOITED_CUSTOM_KEY) : null);
+						List<String> exploitedResourcesUrls =  isStock ?
+								extrasWithoutNamespaces.containsKey(Constants.EXPLOITING_FISHERY_CUSTOM_KEY) ?
+										extrasWithoutNamespaces.get(Constants.EXPLOITING_FISHERY_CUSTOM_KEY) : null:
+											extrasWithoutNamespaces.containsKey(Constants.RESOURCES_EXPLOITED_CUSTOM_KEY) ?
+													extrasWithoutNamespaces.get(Constants.RESOURCES_EXPLOITED_CUSTOM_KEY) : null;
 
 										if(exploitedResourcesUrls != null && !exploitedResourcesUrls.isEmpty()){
 											for (String exploited : exploitedResourcesUrls) {
@@ -298,18 +303,14 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 										// set the values
 										toReturn = new ManageProductBean(
-												semanticId, catalogueIdentifier, uuidKB, grsfType, 
-												grsfDomain, shortName, description, grsfName,traceabilityFlag, 
-												status, recordUrl, sources, similarRecords, 
+												semanticId, catalogueIdentifier, uuidKB, grsfType,
+												grsfDomain, shortName, description, grsfName,traceabilityFlag, sdgFlag,
+												status, recordUrl, sources, similarRecords,
 												connectedBeans, suggestedConnectionsByKnowledgeBase);
 
 			}
 		}
 
-		// save it into session
-		if(toReturn != null)
-			httpSession.setAttribute(sessionProductKey, toReturn);
-
 		logger.debug("Returning item bean " + toReturn);
 		return toReturn;
 	}
@@ -360,7 +361,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 			String username = Utils.getCurrentUser(getThreadLocalRequest()).getUsername();
 
 			// check if the base url of the service is in session
-			String keyPerContext = UtilMethods.concatenateSessionKeyScope(Constants.GRSF_UPDATER_SERVICE, context);
+			String keyPerContext = CatalogueUtilMethods.concatenateSessionKeyScope(Constants.GRSF_UPDATER_SERVICE, context);
 			HttpServletRequest threadRequest = getThreadLocalRequest();
 			String baseUrl = (String)threadRequest.getSession().getAttribute(keyPerContext);
 			if(baseUrl == null ||  baseUrl.isEmpty()){
@@ -372,7 +373,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 			String sessionProductKey = ScopeProvider.instance.get() + bean.getCatalogueIdentifier();
 			threadRequest.getSession().removeAttribute(sessionProductKey);
 
-			Utils.updateRecord(baseUrl, bean, catalogue, username, administratorFullName, threadRequest, 
+			Utils.updateRecord(baseUrl, bean, catalogue, username, administratorFullName, threadRequest,
 					PortalContext.getConfiguration().getCurrentGroupId(threadRequest), context, token);
 
 		}catch(Exception e){
@@ -395,12 +396,13 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 			String baseUrl = "url of the record here";
 			String fullName = "Andrea Rossi";
+			String usernameCurrent = "andrea.rossi";
 			String uuid = UUID.randomUUID().toString();
 			String adminInUrl = "costantino.perciante";
 			String adminInUrlFullName = "Costantino Perciante";
-			long timestamp = System.currentTimeMillis() - 1000 * ((long)(Math.random() * 10 * 60 * 60));
+			long timestamp = System.currentTimeMillis() - 1000 * (long)(Math.random() * 10 * 60 * 60);
 			return new RevertableOperationInfo(
-					baseUrl, fullName, uuid, adminInUrlFullName, adminInUrl, timestamp, RevertableOperations.MERGE);
+					baseUrl, fullName, usernameCurrent, uuid, adminInUrlFullName, adminInUrl, timestamp, RevertableOperations.MERGE);
 		}
 
 		PortalContext pContext = PortalContext.getConfiguration();
@@ -417,7 +419,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 		boolean isReviewer = isReviewer(username, teamRolesByUser);
 
 		if(!(isEditor | isReviewer))
-			throw new Exception("You are not allowed to perform this operation!");
+			throw new Exception("You are not allowed to perform this operation. You must be an editor or a reviewer!");
 
 		// decrypt the url
 		RevertOperationUrl decryptedUrl = new RevertOperationUrl(encryptedUrl);
@@ -434,19 +436,26 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 			throw new Exception("This operation can no longer be reverted (link expired)!");
 
 		DataCatalogue catalogue = getCatalogue(context);
-		String recordUrl = catalogue.getDataset(uuid, catalogue.getApiKeyFromUsername(username)).getExtrasAsHashMap().get(Constants.ITEM_URL_FIELD);
+		CkanDataset dataset = catalogue.getDataset(uuid, catalogue.getApiKeyFromUsername(username));
+		Map<String, String> extras = dataset.getExtrasAsHashMap();
+		String recordUrl = extras.get(Constants.ITEM_URL_FIELD);
+		String currentStatus = extras.get(Constants.STATUS_OF_THE_GRSF_RECORD_CUSTOM_KEY);
+
+		// check current record status
+		if(!currentStatus.equals(Status.To_be_Merged.getOrigName()))
+			throw new Exception("Record '" + dataset.getTitle()  + "' (" + recordUrl + ") is no longer involved in a merge operation!");
 
 		// check if it is a reviewer, than he can do what he wants (no matter the admin)
 		if(isReviewer){
-			return new RevertableOperationInfo(recordUrl, 
-					fullName, uuid, fullNameadminInUrl, userNameadminInUrl, decryptedUrl.getTimestamp(), decryptedUrl.getOperation());
+			return new RevertableOperationInfo(recordUrl,
+					fullName, username, uuid, fullNameadminInUrl, userNameadminInUrl, decryptedUrl.getTimestamp(), decryptedUrl.getOperation());
 		}else{
 
 			if(!username.equals(userNameadminInUrl))
 				throw new Exception("You are not the editor allowed to perform this operation!");
 			else
-				return new RevertableOperationInfo(recordUrl, 
-						fullName, uuid, fullNameadminInUrl, userNameadminInUrl, decryptedUrl.getTimestamp(), decryptedUrl.getOperation());
+				return new RevertableOperationInfo(recordUrl,
+						fullName, username, uuid, fullNameadminInUrl, userNameadminInUrl, decryptedUrl.getTimestamp(), decryptedUrl.getOperation());
 		}
 
 	}
@@ -475,7 +484,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 		try(CloseableHttpClient httpClient = HttpClientBuilder.create().build();){
 
-			String keyPerContext = UtilMethods.concatenateSessionKeyScope(Constants.GRSF_UPDATER_SERVICE, context);
+			String keyPerContext = CatalogueUtilMethods.concatenateSessionKeyScope(Constants.GRSF_UPDATER_SERVICE, context);
 			String baseUrl = (String)getThreadLocalRequest().getSession().getAttribute(keyPerContext);
 			if(baseUrl == null ||  baseUrl.isEmpty()){
 				baseUrl = GRSFUpdaterServiceClient.discoverEndPoint(context);
@@ -485,7 +494,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 			if(baseUrl == null ||  baseUrl.isEmpty())
 				throw new Exception("Unable to discover grsf-updater service!");
 
-			Utils.revertOperation(httpClient, baseUrl, threadRequest, rInfo, token, context, 
+			Utils.revertOperation(httpClient, baseUrl, threadRequest, rInfo, token, context,
 					PortalContext.getConfiguration().getCurrentGroupId(threadRequest));
 
 		}
@@ -522,7 +531,7 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 	private boolean isReviewer(String username, List<GCubeTeam> teamRolesByUser){
 
 		for (GCubeTeam team : teamRolesByUser) {
-			if(team.getTeamName().equals(team.getTeamName().equals(Constants.GRSF_CATALOGUE_REVIEWER_ROLE))){
+			if(team.getTeamName().equals(Constants.GRSF_CATALOGUE_REVIEWER_ROLE)){
 				logger.info("User " + username + " is allowed to modify GRSF records as reviewer");
 				return true;
 			}
@@ -548,10 +557,9 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 
 	@Override
 	public String checkIdentifierExistsInDomain(String id,
-			String domain) throws Exception {
+			String acceptedDomain) throws Exception {
 
 		if(!Utils.isIntoPortal()){
-
 			boolean throwException = Math.random() > 0.5;
 
 			// simulate some delay...
@@ -561,7 +569,6 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 				throw new Exception("The suggested record is not a GRSF record");
 
 			return "http://data.d4science.org/catalogue/grsf_admin/" + id;
-
 		}
 
 		String scopePerCurrentUrl = Utils.getScopeFromClientUrl(getThreadLocalRequest());
@@ -572,19 +579,18 @@ public class GRSFNotificationService extends RemoteServiceServlet implements GRS
 		if(dataset == null)
 			throw new Exception("A record with id " + id  + " doesn't exist");
 
-		if(!dataset.getOrganization().getName().equals(Constants.GRSF_ADMIN_ORGANIZATION_NAME))
-			throw new Exception("The suggested record is not a GRSF record");
+		Map<String, String> extras = dataset.getExtrasAsHashMap();
+		String systemType = extras.get(Constants.SYSTEM_TYPE_CUSTOM_KEY);
+		String domain = extras.get(Constants.DOMAIN_CUSTOM_KEY);
+		String url = extras.get(Constants.ITEM_URL_FIELD);
 
-		List<CkanPair> extrasAsPairs = dataset.getExtras();
+		if(systemType.equals(Constants.SYSTEM_TYPE_FOR_SOURCES_VALUE))
+			throw new Exception("This record is not a GRSF record!");
 
-		for (CkanPair ckanPair : extrasAsPairs) {
-			if(ckanPair.getKey().contains(Constants.DOMAIN_CUSTOM_KEY)){
-				if(ckanPair.getValue().equalsIgnoreCase(domain))
-					return dataset.getExtrasAsHashMap().get(Constants.ITEM_URL_FIELD);
-			}
-		}
+		if(!acceptedDomain.equalsIgnoreCase(domain))
+			throw new Exception("You are suggesting a " + domain + " record instead of a " +  acceptedDomain + " record!");
 
-		throw new Exception("A GRSF record with id " + id  + " doesn't exist in domain " + domain);
+		return url;
 	}
 
 }
